use crate::ast::{Identifier, UseStatement};
use std::fmt::Display;
use std::range::Range;

#[derive(Clone, Debug)]
pub struct SourceDefinition {
    file_id: usize,
    range: Range<usize>,
}

impl SourceDefinition {
    pub fn from_identifier(identifier: &Identifier) -> Self {
        SourceDefinition {
            file_id: identifier.file_id,
            range: identifier.range,
        }
    }

    pub fn from_use_statement(use_statement: &UseStatement) -> Self {
        SourceDefinition {
            file_id: use_statement.file_id,
            range: use_statement.range,
        }
    }

    pub fn file_id(&self) -> usize {
        self.file_id
    }

    pub fn range(&self) -> Range<usize> {
        self.range.clone()
    }
}

#[derive(Debug, Clone)]
pub enum Symbol {
    Function(FunctionSymbol),
    Variable(VariableSymbol),
    Module(ModuleSymbol),
    Class(ClassSymbol),
    UseStatement(UseStatementSymbol),
}

impl Symbol {
    pub fn name(&self) -> &str {
        match self {
            Symbol::Function(s) => s.declared_name.as_str(),
            Symbol::Variable(s) => s.name.as_str(),
            Symbol::Module(s) => s.declared_name.as_str(),
            Symbol::Class(s) => s.declared_name.as_str(),
            Symbol::UseStatement(s) => s.declared_name.as_str(),
        }
    }

    pub fn definition(&self) -> &SourceDefinition {
        match self {
            Symbol::Function(s) => s.definition(),
            Symbol::Module(s) => s.definition(),
            Symbol::Variable(s) => s.definition(),
            Symbol::Class(s) => s.definition(),
            Symbol::UseStatement(s) => s.definition(),
        }
    }
}

impl Display for Symbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        use Symbol::*;
        match self {
            Function(function_symbol) => write!(f, "{}", function_symbol),
            Variable(variable_symbol) => write!(f, "{}", variable_symbol),
            Module(module_symbol) => write!(f, "{}", module_symbol),
            Class(class_symbol) => write!(f, "{}", class_symbol),
            UseStatement(use_statement_symbol) => write!(f, "{}", use_statement_symbol),
        }
    }
}

#[derive(Debug, Clone)]
pub struct FunctionSymbol {
    pub fqn: String,
    pub declared_name: String,
    pub is_public: bool,
    pub is_platform: bool,
    definition: SourceDefinition,
}

impl FunctionSymbol {
    pub fn new(
        fqn: &str,
        declared_name: &str,
        is_public: bool,
        is_platform: bool,
        identifier: &Identifier,
    ) -> FunctionSymbol {
        FunctionSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            is_public,
            is_platform,
            definition: SourceDefinition::from_identifier(identifier),
        }
    }

    fn definition(&self) -> &SourceDefinition {
        &self.definition
    }
}

impl Display for FunctionSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "FunctionSymbol(fqn = {}, declared_name = {}, is_public = {})",
            self.fqn, self.declared_name, self.is_public
        )
    }
}

#[derive(Debug, Clone)]
pub struct VariableSymbol {
    pub name: String,
    pub is_mutable: bool,
    definition: SourceDefinition,
}

impl VariableSymbol {
    pub fn new(name: &str, is_mutable: bool, definition: SourceDefinition) -> Self {
        VariableSymbol {
            name: name.to_string(),
            is_mutable,
            definition,
        }
    }

    pub fn definition(&self) -> &SourceDefinition {
        &self.definition
    }
}

impl Display for VariableSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "VariableSymbol(name = {}, is_mutable = {})",
            self.name, self.is_mutable
        )
    }
}

#[derive(Debug, Clone)]
pub struct ModuleSymbol {
    pub fqn: String,
    pub declared_name: String,
    pub is_public: bool,
    definition: SourceDefinition,
}

impl ModuleSymbol {
    pub fn new(
        fqn: &str,
        declared_name: &str,
        is_public: bool,
        identifier: &Identifier,
    ) -> ModuleSymbol {
        ModuleSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            is_public,
            definition: SourceDefinition::from_identifier(identifier),
        }
    }

    pub fn definition(&self) -> &SourceDefinition {
        &self.definition
    }
}

impl Display for ModuleSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "ModuleSymbol(name = {}, is_public = {})",
            self.fqn, self.is_public
        )
    }
}

#[derive(Debug, Clone)]
pub struct ClassSymbol {
    pub fqn: String,
    pub declared_name: String,
    pub is_public: bool,
    definition: SourceDefinition,
}

impl ClassSymbol {
    pub fn new(fqn: &str, declared_name: &str, is_public: bool, identifier: &Identifier) -> Self {
        ClassSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            is_public,
            definition: SourceDefinition::from_identifier(identifier),
        }
    }

    pub fn definition(&self) -> &SourceDefinition {
        &self.definition
    }
}

impl Display for ClassSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "ClassSymbol(fqn = {}, declared_name = {})",
            self.fqn, self.declared_name
        )
    }
}

#[derive(Debug, Clone)]
pub struct UseStatementSymbol {
    pub fqn: String,
    pub declared_name: String,
    definition: SourceDefinition,
    referenced_symbol: Option<Box<Symbol>>
}

impl UseStatementSymbol {
    pub fn new(fqn: &str, declared_name: &str, definition: SourceDefinition) -> Self {
        UseStatementSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            definition,
            referenced_symbol: None,
        }
    }

    pub fn definition(&self) -> &SourceDefinition {
        &self.definition
    }
    
    pub fn set_referenced_symbol(&mut self, referenced_symbol: Symbol) {
        self.referenced_symbol = Some(Box::new(referenced_symbol));
    }
    
    pub fn referenced_symbol(&self) -> Option<Box<Symbol>> {
        self.referenced_symbol.clone()
    }
}

impl Display for UseStatementSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "UseStatementSymbol(fqn = {}, declared_name = {})",
            self.fqn, self.declared_name
        )
    }
}
