use crate::name_analysis::symbol::Symbol;
use crate::name_analysis::symbol_table::SymbolInsertError::SymbolAlreadyDefined;
use crate::name_analysis::symbol_table::SymbolLookupError::NoDefinition;
use std::collections::HashMap;
use std::fmt::Display;

#[derive(Debug, PartialEq, Clone)]
pub enum ScopeLevel {
    Global,
    ModuleOrInterfaceOrClass,
    Function,
}

#[derive(Debug)]
pub struct Scope {
    level: ScopeLevel,
    parent: Option<usize>,
    use_statement_symbols: HashMap<String, Symbol>,
    module_symbols: HashMap<String, Symbol>,
    type_symbols: HashMap<String, Symbol>,
    function_symbols: HashMap<String, Symbol>,
    variable_symbols: HashMap<String, Symbol>,
    debug_name: String,
}

impl Scope {
    pub fn new(parent: Option<usize>, debug_name: String, level: ScopeLevel) -> Scope {
        Scope {
            level,
            parent,
            use_statement_symbols: HashMap::new(),
            module_symbols: HashMap::new(),
            type_symbols: HashMap::new(),
            function_symbols: HashMap::new(),
            variable_symbols: HashMap::new(),
            debug_name,
        }
    }

    pub fn level(&self) -> ScopeLevel {
        self.level.clone()
    }

    fn get_any_symbol(&self, name: &str) -> Option<&Symbol> {
        self.variable_symbols
            .get(name)
            .or_else(|| self.function_symbols.get(name))
            .or_else(|| self.type_symbols.get(name))
            .or_else(|| self.module_symbols.get(name))
            .or_else(|| self.use_statement_symbols.get(name))
    }

    fn get_usable_symbol(&self, fqn: &str) -> Option<&Symbol> {
        for function_symbol in self.function_symbols.values() {
            match function_symbol {
                Symbol::Function(s) => {
                    if s.fqn == fqn {
                        return Some(function_symbol);
                    }
                }
                _ => panic!("Cannot have non-function in function_symbols."),
            }
        }
        for type_symbol in self.type_symbols.values() {
            match type_symbol {
                Symbol::Class(s) => {
                    if s.fqn == fqn {
                        return Some(type_symbol);
                    }
                }
                _ => panic!("Cannot have non-class in type_symbols."),
            }
        }
        None
    }

    pub fn symbols(&self) -> Vec<&Symbol> {
        let mut symbols: Vec<&Symbol> = Vec::new();
        symbols.extend(self.use_statement_symbols.values());
        symbols.extend(self.module_symbols.values());
        symbols.extend(self.type_symbols.values());
        symbols.extend(self.function_symbols.values());
        symbols.extend(self.variable_symbols.values());
        symbols
    }
}

pub enum SymbolInsertError {
    SymbolAlreadyDefined(Symbol),
}

pub enum SymbolLookupError {
    NoDefinition,
}

#[derive(Debug)]
pub struct SymbolTable {
    scopes: Vec<Scope>,
    current_scope_id: usize,
}

/// Contains a vec of scopes, like a flattened tree
impl SymbolTable {
    pub fn new() -> Self {
        let mut t = SymbolTable {
            scopes: vec![Scope::new(
                None,
                String::from("GlobalScope"),
                ScopeLevel::Global,
            )],
            current_scope_id: 0,
        };
        t
    }

    pub fn current_scope_id(&self) -> usize {
        self.current_scope_id
    }

    pub fn scopes(&self) -> &Vec<Scope> {
        &self.scopes
    }

    pub fn push_scope(&mut self, debug_name: &str, level: ScopeLevel) {
        let id = self.scopes.len();
        self.scopes.push(Scope::new(
            Some(self.current_scope_id),
            debug_name.to_string(),
            level,
        ));
        self.current_scope_id = id;
    }

    pub fn pop_scope(&mut self) {
        if let Some(parent_id) = self.scopes[self.current_scope_id].parent {
            self.current_scope_id = parent_id;
        }
    }

    pub fn insert(&mut self, name: &str, symbol: Symbol) -> Result<(), SymbolInsertError> {
        match symbol {
            Symbol::Function(_) | Symbol::Variable(_) => {
                self.insert_function_or_variable(name, symbol)
            }
            Symbol::Module(_) | Symbol::Class(_) => self.insert_module_or_type(name, symbol),
            Symbol::UseStatement(_) => self.insert_use_statement_symbol(name, symbol),
        }
    }

    fn insert_function_or_variable(
        &mut self,
        name: &str,
        symbol: Symbol,
    ) -> Result<(), SymbolInsertError> {
        if let Some(defined_symbol) = self.scopes[self.current_scope_id]
            .function_symbols
            .get(name)
        {
            Err(SymbolAlreadyDefined(defined_symbol.clone()))
        } else {
            self.scopes[self.current_scope_id]
                .function_symbols
                .insert(name.to_string(), symbol);
            Ok(())
        }
    }

    fn insert_module_or_type(
        &mut self,
        name: &str,
        symbol: Symbol,
    ) -> Result<(), SymbolInsertError> {
        if let Some(defined_symbol) = self.scopes[self.current_scope_id].type_symbols.get(name) {
            Err(SymbolAlreadyDefined(defined_symbol.clone()))
        } else {
            self.scopes[self.current_scope_id]
                .type_symbols
                .insert(name.to_string(), symbol);
            Ok(())
        }
    }

    fn insert_use_statement_symbol(
        &mut self,
        name: &str,
        symbol: Symbol,
    ) -> Result<(), SymbolInsertError> {
        if let Some(defined_symbol) = self.scopes[self.current_scope_id].get_any_symbol(name) {
            Err(SymbolAlreadyDefined(defined_symbol.clone()))
        } else {
            self.scopes[self.current_scope_id]
                .use_statement_symbols
                .insert(name.to_string(), symbol);
            Ok(())
        }
    }

    pub fn lookup(&self, name: &str, scope_id: usize) -> Result<&Symbol, SymbolLookupError> {
        let mut scope_opt = Some(&self.scopes[scope_id]);
        while let Some(scope) = scope_opt {
            if let Some(symbol) = scope.get_any_symbol(name) {
                return Ok(symbol);
            }
            scope_opt = if let Some(parent_id) = scope.parent {
                Some(&self.scopes[parent_id])
            } else {
                None
            };
        }
        Err(NoDefinition)
    }

    pub fn lookup_usable(
        &self,
        fully_qualified_name: &str,
        scope_id: usize,
    ) -> Result<&Symbol, SymbolLookupError> {
        for scope in &self.scopes {
            if scope.level == ScopeLevel::Function {
                continue;
            }
            if let Some(symbol) = scope.get_usable_symbol(fully_qualified_name) {
                return Ok(symbol);
            }
        }
        Err(NoDefinition)
    }
}

impl Display for SymbolTable {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        writeln!(f, "SymbolTable(current_scope = {})", self.current_scope_id)?;
        for (i, scope) in self.scopes.iter().enumerate() {
            writeln!(f, "Scope {} {}", i, scope.debug_name)?;
            for (name, symbol) in &scope.use_statement_symbols {
                writeln!(f, "  {}({})", name, symbol)?;
            }
            for (name, symbol) in &scope.module_symbols {
                writeln!(f, "  {}({})", name, symbol)?;
            }
            for (name, symbol) in &scope.type_symbols {
                writeln!(f, "  {}({})", name, symbol)?;
            }
            for (name, symbol) in &scope.function_symbols {
                writeln!(f, "  {}({})", name, symbol)?;
            }
            for (name, symbol) in &scope.variable_symbols {
                writeln!(f, "  {}({})", name, symbol)?;
            }
        }
        Ok(())
    }
}
